from coverage_strategies.coverage_strategies.Entities import Slot, Strategy


class CircleInsideFromCornerFarthestFromIo_Strategy(Strategy):
    def get_steps(self, agent_r, board_size = 50, agent_o = None):
        """
            This function return the agent steps, when deciding to cover the world, spiraling from outside to inside.
            Note: this coverage method is not optimal!
            :param agent: the agent covering the world
            :param board_size: self explanatory
            :return: list of steps
            """

        next_slot = Slot(agent_r.InitPosX, agent_r.InitPosY)
        # todo: impreve performance by using extending of lists
        # start by going toward the closest corner
        if next_slot.row < board_size / 2:
            while next_slot.row > 0:
                self.steps.append(next_slot)
                next_slot = next_slot.go_north()
                continue
        else:
            while next_slot.row < board_size - 1:
                self.steps.append(next_slot)
                next_slot = next_slot.go_south()
                continue

        if next_slot.col < board_size / 2:
            while next_slot.col > 0:
                self.steps.append(next_slot)
                next_slot = next_slot.go_west()
                continue
        else:
            while next_slot.col < board_size - 1:
                self.steps.append(next_slot)
                next_slot = next_slot.go_east()
                continue
        self.steps.append(next_slot)

        # after reaching the closest-to-start corner, start covering the world, counter clockwise
        shallow_slots = [[0 for x in range(board_size)] for y in range(board_size)]
        shallow_slots[next_slot.row][next_slot.col] = 1
        dist_from_edge = 0

        while dist_from_edge < board_size / 2:
            if next_slot.row + dist_from_edge < board_size - 1 and next_slot.col == dist_from_edge:
                direction = 's'
            elif next_slot.row + dist_from_edge == board_size - 1 and next_slot.col + dist_from_edge < board_size - 1:
                direction = 'e'
            elif next_slot.row > dist_from_edge and next_slot.col + dist_from_edge == board_size - 1:
                direction = 'n'
            elif next_slot.row == dist_from_edge and next_slot.col + dist_from_edge >= board_size - 1:
                direction = 'w'

            if direction == 's':
                new_slot = next_slot.go_south()
            elif direction == 'e':
                new_slot = next_slot.go_east()
            elif direction == 'n':
                new_slot = next_slot.go_north()
            elif direction == 'w':
                new_slot = next_slot.go_west()

            if shallow_slots[new_slot.row][new_slot.col] == 1:
                dist_from_edge += 1
                continue
            else:
                next_slot = new_slot

            self.steps.append(next_slot)
            shallow_slots[next_slot.row][next_slot.col] = 1

        return self.steps


from coverage_strategies.coverage_strategies import CircleOutsideFromCornerAdjacentToIo_Strategy
from coverage_strategies.coverage_strategies.CircleInsideFromCornerFarthestFromIo_Strategy import \
    CircleInsideFromCornerFarthestFromIo_Strategy
from coverage_strategies.coverage_strategies.CircleOutsideFromBoardCenter_Strategy import CircleOutsideFromBoardCenter_Strategy
from coverage_strategies.coverage_strategies.CircleOutsideFromCornerFarthestFromIo_Strategy import \
    CircleOutsideFromCornerFarthestFromIo_Strategy
from coverage_strategies.coverage_strategies.CircleOutsideFromIo_Strategy import CircleOutsideFromIo_Strategy
from coverage_strategies.coverage_strategies.CoverByQuarters_Strategy import CoverByQuarters_Strategy
from coverage_strategies.coverage_strategies.Entities import StrategyEnum
from coverage_strategies.coverage_strategies.HorizontalCircularCoverage_Strategy import HorizontalCircularCoverage_Strategy
from coverage_strategies.coverage_strategies.InterceptThenCopy_Strategy import InterceptThenCopy_Strategy
from coverage_strategies.coverage_strategies.LCP_Strategy import LCP_Strategy
from coverage_strategies.coverage_strategies.LongestToReach_Strategy import LongestToReach_Strategy
from coverage_strategies.coverage_strategies.STC_Strategy import STC_Strategy
from coverage_strategies.coverage_strategies.TrulyRandom_Strategy import TrulyRandomStrategy
from coverage_strategies.coverage_strategies.VerticalCircularCoverage_Strategy import VerticalCircularCoverage_Strategy
from coverage_strategies.coverage_strategies.VerticalCoverageFromCornerFarthestFromIo_Strategy import \
    VerticalCoverageFromCornerFarthestFromIo_Strategy
from coverage_strategies.coverage_strategies.VerticalNonCircularCoverage_Strategy import VerticalNonCircularCoverage_Strategy


def get_strategy_from_enum(strategy_enum: StrategyEnum):
    if strategy_enum == StrategyEnum.VerticalCoverageCircular:
        return VerticalCircularCoverage_Strategy()
    elif strategy_enum == StrategyEnum.HorizontalCoverageCircular:
        return HorizontalCircularCoverage_Strategy()
    elif strategy_enum == StrategyEnum.FullKnowledgeInterceptionCircular:
        return InterceptThenCopy_Strategy()
    elif strategy_enum == StrategyEnum.QuartersCoverageCircular:
        return CoverByQuarters_Strategy()
    elif strategy_enum == StrategyEnum.RandomSTC:
        return STC_Strategy()
    elif strategy_enum == StrategyEnum.VerticalCoverageNonCircular:
        return VerticalNonCircularCoverage_Strategy()
    elif strategy_enum == StrategyEnum.SpiralingIn:
        return CircleInsideFromCornerFarthestFromIo_Strategy()
    elif strategy_enum == StrategyEnum.SpiralingOut:
        return CircleOutsideFromBoardCenter_Strategy()
    elif strategy_enum == StrategyEnum.VerticalFromFarthestCorner_OpponentAware:
        return VerticalCoverageFromCornerFarthestFromIo_Strategy()
    elif strategy_enum == StrategyEnum.SemiCyclingFromFarthestCorner_OpponentAware:
        return CircleOutsideFromCornerFarthestFromIo_Strategy()
    elif strategy_enum == StrategyEnum.SemiCyclingFromAdjacentCorner_col_OpponentAware:
        return CircleOutsideFromCornerAdjacentToIo_Strategy.CircleOutsideFromCornerAdjacentToIo_Strategy(False)
    elif strategy_enum == StrategyEnum.SemiCyclingFromAdjacentCorner_row_OpponentAware:
        return CircleOutsideFromCornerAdjacentToIo_Strategy.CircleOutsideFromCornerAdjacentToIo_Strategy(True)
    elif strategy_enum == StrategyEnum.CircleOutsideFromIo:
        return CircleOutsideFromIo_Strategy()
    elif strategy_enum == StrategyEnum.LCP:
        return LCP_Strategy()
    elif strategy_enum == StrategyEnum.LONGEST_TO_REACH:
        return LongestToReach_Strategy()
    elif strategy_enum == StrategyEnum.TRULY_RANDOM:
        return TrulyRandomStrategy()


from abc import ABCMeta
from abc import abstractmethod
from enum import Enum
import numpy as np

# from coverage_strategies.StrategyGenerator import get_strategy_from_enum


class Board:
    def __init__(self, rows, cols):
        self.Rows = rows
        self.Cols = cols
        self.Slots = [[Slot(y, x) for x in range(self.Rows)] for y in range(self.Cols)]
        self.Obstacles = []

    def __str__(self):
        s = ''
        for i in range(len(self.Slots)):
            for j in range(len(self.Slots[i])):
                s += str(self.Slots[i][j])
            s += '\n'
        return s

    def reset(self):
        for s in self.Slots:
            s.has_been_visited = False
            s.is_occupied = False
            s.covered_by = "*"

    def add_obstacles(self,obs=[]):
        self.Obstacles.extend(obs)

class Slot:
    def __init__(self, x, y):
        self.has_been_visited = False
        self.covered_by = "*"
        self.row = x
        self.col = y
        self.Name = "("+str(x) + "," + str(y) + ")"

    def __eq__(self, other):
        return self.row == other.row and self.col == other.col

    def __ne__(self, other):
        return self.row != other.row or self.col != other.col

    def __hash__(self):
        return hash((self.row, self.col))

    def go_west(self):
        return Slot(self.row, self.col - 1)

    def go_east(self):
        return Slot(self.row, self.col + 1)

    def go_north(self):
        return Slot(self.row - 1, self.col)

    def go_south(self):
        return Slot(self.row + 1, self.col)

    def go(self, s, opposite_direction = False):
        """
        Go South, East, North or West, according to the given parameter.
        :param s: Must by 'd', 'r','n','l'
        :param opposite_direction: False by default. Move to the opposite direction
        :return: the new slot
        """
        assert s == 'u' or s == 'd' or s == 'r' or s == 'l'

        if opposite_direction:
            if s == 'u':
                s = 'd'
            if s == 'd':
                s = 'u'
            if s == 'r':
                s = 'l'
            if s == 'l':
                s = 'r'

        if s == 'u':
            return self.go_north()
        if s == 'd':
            return self.go_south()
        if s == 'r':
            return self.go_east()
        if s == 'l':
            return self.go_west()

    def to_tuple(self):
        return self.row, self.col

    def __str__(self):
        return "({0},{1})".format(str(int(self.row)), str(int(self.col)))

    def __repr__(self):
        return str(self)


class StrategyEnum(Enum):
    VerticalCoverageCircular = 0
    HorizontalCoverageCircular = 1
    FullKnowledgeInterceptionCircular = 2
    QuartersCoverageCircular = 3
    RandomSTC = 4
    VerticalCoverageNonCircular = 5
    SpiralingOut = 6
    SpiralingIn = 7
    VerticalFromFarthestCorner_OpponentAware = 8
    SemiCyclingFromFarthestCorner_OpponentAware = 9
    CircleOutsideFromIo = 10,
    LCP = 11,
    LONGEST_TO_REACH = 12,
    TRULY_RANDOM = 13,
    SemiCyclingFromAdjacentCorner_col_OpponentAware = 14,
    SemiCyclingFromAdjacentCorner_row_OpponentAware = 15


class Agent:
    def __init__(self, name: str, strategy_enum: StrategyEnum, x: int, y: int, board: Board = None,
                 agent_o: object = None) -> None:

        assert isinstance(strategy_enum, Enum)

        self.Name = name
        self.StrategyEnum = strategy_enum
        self.InitPosX = x
        self.InitPosY = y
        self.gameBoard = board

        from coverage_strategies.coverage_strategies.StrategyGenerator import get_strategy_from_enum
        self.Strategy = get_strategy_from_enum(strategy_enum)
        self.steps = self.Strategy.get_steps(self, len(board.Slots), agent_o)

    def get_tdv(self):
        from math import fabs
        return sum([(1/(1+stepI))*(fabs(self.steps[stepI].row - self.steps[0].row)+fabs(self.steps[stepI].col - self.steps[0].col)) for stepI in range(len(self.steps))])

    def get_frame(self):
        pass


    def get_strategy(self):
        return self.Strategy.__str__()

    def get_strategy_short(self):
        return self.get_strategy()[:5] + "..."

    def display_heat_map(self,x,y):
        arr = self.get_heatmap()
        # DisplayingClass.create_heat_map(arr, x, y, self.get_strategy_short())

    def get_heatmap(self):
        arr = np.zeros((self.gameBoard.Rows, self.gameBoard.Cols))
        for id in range(len(self.steps)):
            if arr[self.steps[id].row][self.steps[id].col] == 0:
                arr[self.steps[id].row][self.steps[id].col] = id
        return arr

    def get_cross_heatmap(self, other, probabilities=[0.5, 0.5]):
        my_hm = probabilities[0] * self.get_heatmap()
        o_hm = probabilities[1] * other.get_heatmap()
        return np.add(my_hm, o_hm)

    def get_sub_heatmap(self, other_hm, probabilities=[0.5, 0.5]):
        my_hm = probabilities[0] * self.get_heatmap()
        o_hm = probabilities[1] * other_hm
        return np.subtract(my_hm, o_hm)

    # def display_cross_heatmap(self, other, display_grid_x, display_grid_y, probabilities):
    #     c = self.get_cross_heatmap(other, probabilities)
    #     DisplayingClass.create_heat_map(c, display_grid_x, display_grid_y,
    #                                     "comb. of \n({0} and \n{1}):".format(
    #                                         str(self.get_strategy_short()), str(other.get_strategy_short())))
    #     return c

    # def display_sub_heatmap(self, other_hm, display_grid_x, display_grid_y, probabilities):
    #     c = self.get_sub_heatmap(other_hm, probabilities)
    #     #
    #     # DisplayingClass.create_heat_map(c, display_grid_x, display_grid_y,
    #     #                                 "subs. sum: {}\navg: {}\n std: {}".format(np.sum(c),
    #     #                                                                           np.average(c),
    #     #                                                                           np.std(c)))

class Game:
    def __init__(self, agent_r: Agent, agent_o: Agent, size=(100,100)) -> None:
        self._board = agent_r.gameBoard
        self._agentR = agent_r
        self._agentO = agent_o

    def run_game(self, enforce_paths_length=False):
        steps_r = self._agentR.steps
        steps_o = self._agentO.steps

        if enforce_paths_length:
            if not len(steps_o) == len(steps_r):
                raise AssertionError("wrong length! len(steps_o)={}, len(steps_r)={}".format(len(steps_o),len(steps_r)))

        for i in range(min(len(steps_r), len(steps_o))):
            # perform step for R
            step_r = steps_r[i]
            if not self._board.Slots[int(step_r.row)][int(step_r.col)].has_been_visited:
                self._board.Slots[int(step_r.row)][int(step_r.col)].has_been_visited = True
                self._board.Slots[int(step_r.row)][int(step_r.col)].covered_by = self._agentR.Name

            # then perform step for O
            step_o = steps_o[i]
            if not self._board.Slots[int(step_o.row)][int(step_o.col)].has_been_visited:
                self._board.Slots[int(step_o.row)][int(step_o.col)].has_been_visited = True
                self._board.Slots[int(step_o.row)][int(step_o.col)].covered_by = self._agentO.Name

        return self.get_r_gain(), self.get_o_gain()
    def get_r_gain(self):
        cond_count = 0

        # print self._board.Slots

        for i in range(0, self._board.Rows):
            for j in range(0, self._board.Cols):
                if self._board.Slots[i][j].covered_by == self._agentR.Name:
                    cond_count += 1

        return float(cond_count)

    def get_o_gain(self):
        cond_count = 0

        size_x = len(self._board.Slots)
        size_y = len(self._board.Slots[0])

        for i in range(0, size_x):
            for j in range(0, size_y):
                if self._board.Slots[i][j].covered_by == self._agentO.Name:
                    cond_count += 1

        return float(cond_count)

    @property
    def board(self):
        return self._board


class Strategy:
    __metaclass__ = ABCMeta
    steps = [] # type: List[Slot]

    def __init__(self):
        self.steps = []
        self.set_steps = set()

    def __str__(self):
        return self.__class__.__name__

    @classmethod
    @abstractmethod
    def get_steps(self, agent_r, board_size = 50, agent_o = None):
        """Returns the steps agent perform to cover the world"""

    @classmethod
    def go_from_a_to_b(self, a, b):
        """
        Returns a list of steps from A to B
        :param a: First Slot
        :param b: Second Slot
        :return: list of steps from A to B
        """

        current_step = a
        steps_to_return = [current_step]

        while not current_step.row == b.row:
            current_step = current_step.go_north() if b.row < a.row else current_step.go_south()
            steps_to_return.append(current_step)

        while not current_step.col == b.col:
            current_step = current_step.go_east() if b.col > a.col else current_step.go_west()
            steps_to_return.append(current_step)

        return steps_to_return

    @classmethod
    def get_farthest_corner(self, a, board_size):
        """
        return the farthest corner from a given position
        :param a: the given position
        :param board_size: the size of the given game board
        :return: the farthest corner from A
        """
        f_row = 0 if a.row > board_size / 2 else board_size - 1
        f_col = 0 if a.col > board_size / 2 else board_size - 1
        return Slot(f_row, f_col)




from coverage_strategies.coverage_strategies.Entities import Slot, Strategy


class CircleOutsideFromCornerAdjacentToIo_Strategy(Strategy):
    first_option = None

    def __init__(self, first_option):
        super().__init__()
        self.first_option = first_option

    def get_steps(self, agent_r, board_size = 50, agent_o = None):
        """
        This function returns the coverage self.steps, when covering knowing io, starting from the a corner adjacent to
        io, and covering semi-cyclic - covering the closer layers first.
        :param self:
        :param board_size:
        :param agent_o:
        :return: the coverage self.steps.
        """
        assert agent_o is not None
        
        # go to the adjacent corner
        steps_to_start = Strategy.go_from_a_to_b(a=Slot(agent_r.InitPosX, agent_r.InitPosY),
                                                  b=Strategy.get_adjacent_corner(
                                                      a=Slot(agent_o.InitPosX, agent_o.InitPosY),
                                                      board_size=board_size, first_option=self.first_option))
        for i in steps_to_start:
            self.add_step(i)

        # from there, cover semi-cyclic
        current_slot = self.steps[-1]
        v_dir = 'u' if current_slot.row == board_size - 1 else 'd'
        h_dir = 'r' if current_slot.col == board_size - 1 else 'l'
        start_vertical = True
        distance = 1
        counter = 1

        # initial horizontal step
        current_slot = current_slot.go_west() if h_dir == 'r' else current_slot.go_east()
        self.add_step(current_slot)

        counter += 1

        while counter <= board_size * board_size and distance < board_size:
            if start_vertical:
                # going vertically
                for _ in range(distance):
                    current_slot = current_slot.go_north() if v_dir == 'u' else current_slot.go_south()
                    self.add_step(current_slot)
                    counter += 1

                # going horizontally
                for _ in range(distance):
                    current_slot = current_slot.go_west() if h_dir == 'l' else current_slot.go_east()
                    self.add_step(current_slot)
                    counter += 1

                # final vertical step
                if counter < board_size * board_size:
                    current_slot = current_slot.go_north() if v_dir == 'u' else current_slot.go_south()
                    self.add_step(current_slot)
                    counter += 1

            else:
                # going horizontally
                for _ in range(distance):
                    current_slot = current_slot.go_west() if h_dir == 'l' else current_slot.go_east()
                    self.add_step(current_slot)
                    counter += 1

                # going vertically
                for _ in range(distance):
                    current_slot = current_slot.go_north() if v_dir == 'u' else current_slot.go_south()
                    self.add_step(current_slot)
                    counter += 1

                # final horizontal step
                if counter < board_size * board_size:
                    current_slot = current_slot.go_west() if h_dir == 'l' else current_slot.go_east()
                    self.add_step(current_slot)
                    counter += 1

            start_vertical = not start_vertical
            h_dir = 'r' if h_dir == 'l' else 'l'
            v_dir = 'u' if v_dir == 'd' else 'd'

            distance += 1

        return self.steps




from abc import abstractmethod
from coverage_strategies import Slot, Strategy, Agent


class GrowingDiagonalsCornerToCorner_Strategy(Strategy):
    @abstractmethod
    def get_steps(self, agent_r: Agent, board_size=50, agent_o: Agent = None):
        assert agent_o is not None

        # assert o's initial position is a corner
        o_init_slot = Slot(agent_o.InitPosX, agent_o.InitPosY)
        assert o_init_slot == (0, 0) or \
               o_init_slot == (0, board_size - 1) or \
               o_init_slot == (board_size - 1, 0) or \
               o_init_slot == (board_size - 1, board_size - 1)

        # go to the farthest corner
        steps_to_start = Strategy.go_from_a_to_b(a=Slot(agent_r.InitPosX, agent_r.InitPosY),
                                                 b=Strategy.get_adjacent_corner(
                                                     a=Slot(agent_o.InitPosX, agent_o.InitPosY),
                                                     board_size=board_size))  # or false!
        for i in steps_to_start:
            self.add_step(i)

        advance_counter = 1

        return self.steps



from coverage_strategies.coverage_strategies.Entities import Strategy, Slot, Board, Agent
from coverage_strategies.coverage_strategies.SpanningTreeCoverage import is_slot_shallow_obstacle


def get_graph_from_board(b:Board):
    g = Graph()

    for cell_row in b.Slots:
        for cell in cell_row:
            if g.get_vertex(cell) is None:
                g.add_vertex(cell)
            for n in [cell.go_north(), cell.go_south(), cell.go_east(), cell.go_west()]:
                if 0 <= n.row < b.Rows and 0 <= n.col < b.Cols:
                    g.add_edge(cell, n, 1 if not (is_slot_shallow_obstacle(cell, b.Obstacles) or is_slot_shallow_obstacle(n,b.Obstacles)) else sys.maxsize)
    return g

def get_interception_point(steps_o, InitPosX, InitPosY, agentO:Agent):
    ip_x = -1
    ip_y = -1

    steps_counter = 0
    g = get_graph_from_board(agentO.gameBoard)
    start = Slot(InitPosX, InitPosY)
    path_to = []
    for step in steps_o:
        steps_counter += 1

        # computing the distance should be done using dijkstra, since now we allow obstacles
        # create graph from Board (with obstacles)

        dijkstra(g, g.get_vertex(start), g.get_vertex(step))

        target = g.get_vertex(step)
        path = [target.get_id()]
        shortest(target, path)
        # print('The shortest path : %s' % (path[::-1]))

        distance_from_i_r = len(path[::-1])
        if steps_counter > distance_from_i_r:
            ip_x, ip_y = step.row, step.col
            path_to = path[::-1]
            break

    assert (ip_x != -1 and ip_y != -1)

    return Slot(ip_x, ip_y), len(path_to), path_to


class InterceptThenCopy_Strategy(Strategy):
    def get_steps(self, agent_r, board_size=50, agent_o=None):
        # steps_o = StrategyGenerator.get_strategy_from_enum(agent_o.StrategyEnum).get_steps(agent_o, board_size,
        #                                                                                    board_size)
        steps_o = agent_o.steps

        # Find interception point
        (interceptionPoint_R_O, distance, path_to_interception_point) = get_interception_point(steps_o, agent_r.InitPosX, agent_r.InitPosY, agent_o)
        steps_r = run_agent_over_board_interception_strategy(steps_o, path_to_interception_point)
        return steps_r


def run_agent_over_board_interception_strategy(steps_o, path_to):
    steps = []
    steps.extend(path_to)
    steps.extend(steps_o[(steps_o.index(path_to[-1])+1):])

    return steps


# --------------------------------------
import sys
from functools import total_ordering

@total_ordering
class Vertex:
    def __init__(self, node):
        self.id = node
        self.adjacent = {}
        # Set distance to infinity for all nodes
        self.distance = sys.maxsize
        # Mark all nodes unvisited
        self.visited = False
        # Predecessor
        self.previous = None

    def add_neighbor(self, neighbor, weight=0):
        self.adjacent[neighbor] = weight

    def get_connections(self):
        return self.adjacent.keys()

    def get_id(self):
        return self.id

    def get_weight(self, neighbor):
        return self.adjacent[neighbor]

    def set_distance(self, dist):
        self.distance = dist

    def get_distance(self):
        return self.distance

    def set_previous(self, prev):
        self.previous = prev

    def set_visited(self):
        self.visited = True

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.distance == other.distance
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self.distance < other.distance

    def __hash__(self):
        return id(self)

    def __str__(self):
        return str(self.id) + ' adjacent: ' + str([x.id for x in self.adjacent])


class Graph:
    def __init__(self):
        self.vert_dict = {}
        self.num_vertices = 0

    def __iter__(self):
        return iter(self.vert_dict.values())

    def add_vertex(self, node):
        self.num_vertices = self.num_vertices + 1
        new_vertex = Vertex(node)
        self.vert_dict[node] = new_vertex
        return new_vertex

    def get_vertex(self, n):
        if n in self.vert_dict:
            return self.vert_dict[n]
        else:
            return None

    def add_edge(self, frm, to, cost=0):
        if frm not in self.vert_dict:
            self.add_vertex(frm)
        if to not in self.vert_dict:
            self.add_vertex(to)

        self.vert_dict[frm].add_neighbor(self.vert_dict[to], cost)
        self.vert_dict[to].add_neighbor(self.vert_dict[frm], cost)

    def get_vertices(self):
        return self.vert_dict.keys()

    def set_previous(self, current):
        self.previous = current

    def get_previous(self, current):
        return self.previous


def shortest(v, path):
    ''' make shortest path from v.previous'''
    if v.previous:
        path.append(v.previous.get_id())
        shortest(v.previous, path)
    return


import heapq


def dijkstra(aGraph, start, target):
    # print('''Dijkstra's shortest path''')
    # Set the distance for the start node to zero
    start.set_distance(0)

    # Put tuple pair into the priority queue
    unvisited_queue = [(v.get_distance(), v) for v in aGraph]
    heapq.heapify(unvisited_queue)

    while len(unvisited_queue):
        # Pops a vertex with the smallest distance
        uv = heapq.heappop(unvisited_queue)
        current = uv[1]
        current.set_visited()

        # for next in v.adjacent:
        for next in current.adjacent:
            # if visited, skip
            if next.visited:
                continue
            new_dist = current.get_distance() + current.get_weight(next)

            if new_dist < next.get_distance():
                next.set_distance(new_dist)
                next.set_previous(current)
                # print('updated : current = %s next = %s new_dist = %s' \
                #       % (current.get_id(), next.get_id(), next.get_distance()))
            else:
                pass
                # print('not updated : current = %s next = %s new_dist = %s' \
                #       % (current.get_id(), next.get_id(), next.get_distance()))

        # Rebuild heap
        # 1. Pop every item
        while len(unvisited_queue):
            heapq.heappop(unvisited_queue)
        # 2. Put all vertices not visited into the queue
        unvisited_queue = [(v.get_distance(), v) for v in aGraph if not v.visited]
        heapq.heapify(unvisited_queue)

# A working example of Graph and dijkstra
# if __name__ == '__main__':
#
#     g = Graph()
#
#     g.add_vertex('a')
#     g.add_vertex('b')
#     g.add_vertex('c')
#     g.add_vertex('d')
#     g.add_vertex('e')
#     g.add_vertex('f')
#
#     g.add_edge('a', 'b', 7)
#     g.add_edge('a', 'c', 9)
#     g.add_edge('a', 'f', 14)
#     g.add_edge('b', 'c', 10)
#     g.add_edge('b', 'd', 15)
#     g.add_edge('c', 'd', 11)
#     g.add_edge('c', 'f', 2)
#     g.add_edge('d', 'e', 6)
#     g.add_edge('e', 'f', 9)
#
#     print('Graph data:')
#     for v in g:
#         for w in v.get_connections():
#             vid = v.get_id()
#             wid = w.get_id()
#             print('( %s , %s, %3d)' % (vid, wid, v.get_weight(w)))
#
#     dijkstra(g, g.get_vertex('a'), g.get_vertex('e'))
#
#     target = g.get_vertex('e')
#     path = [target.get_id()]
#     shortest(target, path)
#     print('The shortest path : %s' % (path[::-1]))



from coverage_strategies.coverage_strategies.Entities import Strategy, Slot
from coverage_strategies.coverage_strategies import SpanningTreeCoverage


class STC_Strategy(Strategy):
    # @abstractmethod
    def get_steps(self, agent_r, board_size = 50, agent_o = None,):
        return SpanningTreeCoverage.get_random_coverage_strategy(size=board_size,
                                                                 i_r=Slot(agent_r.InitPosX, agent_r.InitPosY),
                                                                 print_mst=True,
                                                                 obstacles=agent_r.gameBoard.Obstacles)

